from tfvit.model import ViT
from tfvit.model import ViTTiny16224, ViTTiny16384
from tfvit.model import ViTSmall16224, ViTSmall16384, ViTSmall32224, ViTSmall32384
from tfvit.model import ViTBase16224, ViTBase16384, ViTBase32224, ViTBase32384
from tfvit.model import ViTLarge16224, ViTLarge16384
from tfvit.model import ViTSmall14518, ViTBase14518, ViTLarge14518, ViTGiant14518
